""" Utilities for validating LEMS models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-06-03
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...log.data_model import StandardOutputErrorCapturerLevel
from ...log.utils import StandardOutputErrorCapturer
from lems.model.model import Model
from pyneuroml.pynml import get_path_to_jnml_jar, run_jneuroml, DEFAULTS
import os
import shutil
import tempfile
import zipfile


def validate_model(filename, name=None):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`Model`: model
    """
    errors = []
    warnings = []
    model = None

    with StandardOutputErrorCapturer(relay=False, level=StandardOutputErrorCapturerLevel.python) as captured:
        try:
            validate_neuroml2_lems_file(filename)
        except SystemExit:
            errors.append(['`{}` is not a valid LEMS file.'.format(filename), [[captured.get_text() or '']]])
            return (errors, warnings, model)

    core_types_dir = tempfile.mkdtemp()
    jar_filename = get_path_to_jnml_jar()
    with zipfile.ZipFile(jar_filename, 'r') as jar_file:
        neuroml2_core_type_members = (name for name in jar_file.namelist() if name.startswith('NeuroML2CoreTypes/'))
        jar_file.extractall(core_types_dir, members=neuroml2_core_type_members)

    model = Model(include_includes=True, fail_on_missing_includes=True)
    model.add_include_directory(os.path.join(core_types_dir, 'NeuroML2CoreTypes'))
    model.import_from_file(filename)
    shutil.rmtree(core_types_dir)

    return (errors, warnings, model)


def validate_neuroml2_lems_file(nml2_lems_file_name, max_memory=DEFAULTS['default_java_max_memory']):
    # type (str, str) -> bool
    """Validate a NeuroML 2 LEMS file using jNeuroML.

    Note that this uses jNeuroML and so is aware of the standard NeuroML LEMS
    definitions.

    TODO: allow inclusion of other paths for user-defined LEMS definitions
    (does the -norun option allow the use of -I?)

    :param nml2_lems_file_name: name of file to validate
    :type nml2_lems_file_name: str
    :param max_memory: memory to use for the Java virtual machine
    :type max_memory: str
    :returns: True if valid, False if invalid

    """
    post_args = ""
    post_args += "-norun"

    return run_jneuroml(
        "",
        nml2_lems_file_name,
        post_args,
        max_memory=max_memory,
        verbose=False,
        report_jnml_output=True,
        exit_on_fail=True)



""" Methods for reading specifications of simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..biosimulations.utils import validate_biosimulations_api_response
from ..config import get_config
from ..utils.core import patch_dict
from kisao import Kisao
from kisao.data_model import TermType as KisaoTermType
from kisao.utils import get_term_type as get_kisao_term_type
import json
import os
import requests
import simplejson.errors


def read_simulator_specs(path_or_url, patch=None, validate=True):
    """ Read the specifications of a simulator

    Args:
        path_or_url (:obj:`str`): file path or URL for the specifications of a simulator
        patch (:obj:`dict`, optional): values of properties to supersede those in :obj:`path_or_url`
        validate (:obj:`bool`, optional): whether to validate the specifications

    Returns:
        :obj:`dict`: specifications of a simulator

    Raises:
        :obj:`requests.RequestException`: if the specifications could not be retrieved
        :obj:`simplejson.errors.JSONDecodeError`: if the specifications are not propertly encoded into JSON
        :obj:`ValueError`: if the specifications are not consistent with the BioSimulators schema
    """
    if os.path.isfile(path_or_url):
        with open(path_or_url, 'r') as file:
            try:
                specs = json.load(file)
            except json.JSONDecodeError as error:
                raise ValueError(''.join([
                    'Simulator specifications from {} could not be parsed. '.format(path_or_url),
                    'Specifications must be encoded into JSON.\n\n  {}'.format(str(error).replace('\n', '\n  ')),
                ]))

    else:
        # download specifications
        response = requests.get(path_or_url)
        try:
            response.raise_for_status()
        except requests.RequestException as error:
            raise requests.RequestException('Simulator specifications could not be retrieved from {}.\n\n  {}'.format(
                path_or_url, str(error).replace('\n', '\n  ')))

        # check that specifications is valid JSON
        try:
            specs = response.json()
        except simplejson.errors.JSONDecodeError as error:
            raise ValueError(''.join([
                'Simulator specifications from {} could not be parsed. '.format(path_or_url),
                'Specifications must be encoded into JSON.\n\n  {}'.format(str(error).replace('\n', '\n  ')),
            ]))

    # apply patch
    if patch:
        patch_dict(specs, patch)

    # validate specifications
    if validate:
        api_endpoint = get_config().BIOSIMULATORS_API_ENDPOINT
        response = requests.post('{}simulators/validate'.format(api_endpoint), json=specs)
        intro_failure_msg = ''.join([
            "The simulator specifications from `{}` are invalid. ".format(path_or_url),
            "The specifications of simulation tools must adhere to BioSimulators' schema. ",
            "BioSimulators' schema is available in both JSON Schema and Open API Specifications formats. ",
            "Documentation is available at {}.".format(api_endpoint)
        ])
        validate_biosimulations_api_response(response, intro_failure_msg)

        kisao = Kisao()
        for i_alg, alg_specs in enumerate(specs['algorithms']):
            alg_term = kisao.get_term(alg_specs['kisaoId']['id'])
            alg_term_type = get_kisao_term_type(alg_term)
            if alg_term_type != KisaoTermType.algorithm:
                msg = 'KiSAO term `{}` of algorithm {} must be a term for an algorithm, not a `{}` term.'.format(
                    alg_specs['kisaoId']['id'], i_alg + 1, alg_term_type.name)
                raise ValueError(msg)

            for i_param, param_specs in enumerate(alg_specs['parameters']):
                param_term = kisao.get_term(param_specs['kisaoId']['id'])
                param_term_type = get_kisao_term_type(param_term)
                if param_term_type != KisaoTermType.algorithm_parameter:
                    msg = (
                        'KiSAO term `{}` of parameter {} of algorithm {} must be a term for an algorithm parameter, not a `{}` term.'
                    ).format(param_specs['kisaoId']['id'], i_param + 1, i_alg + 1, param_term_type.name)
                    raise ValueError(msg)

    # return validated specifications
    return specs



""" Standard exceptions issued by biosimulation tools

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

__all__ = [
    'BioSimulatorsException',
]


class BioSimulatorsException(Exception):
    """ Base class for simulator exceptions """
    pass  # pragma: no cover



""" Methods for validating COMBINE archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-16
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config, Config  # noqa: F401
from ..omex_meta.io import read_omex_meta_files_for_archive
from ..sedml.io import SedmlSimulationReader
from .data_model import CombineArchive, CombineArchiveContent, CombineArchiveContentFormat, CombineArchiveContentFormatPattern  # noqa: F401
from .utils import get_sedml_contents
import imghdr
import os
import re

__all__ = [
    'validate',
    'validate_format',
    'validate_content',
]


def validate(archive, archive_dirname,
             include_all_sed_docs_when_no_sed_doc_is_master=True,
             always_include_all_sed_docs=False,
             formats_to_validate=[
                 CombineArchiveContentFormat.SED_ML,
             ],
             validate_models_with_languages=True,
             config=None):
    """ Validate a COMBINE/OMEX archive and the SED-ML and model documents it contains

    Args:
        archive (:obj:`CombineArchive`): COMBINE/OMEX archive
        archive_dirname (:obj:`str`): directory with the content of the archive
        include_all_sed_docs_when_no_sed_doc_is_master (:obj:`bool`, optional): if :obj:`true`
            and no SED document has ``master="true"``, return all SED documents.
        always_include_all_sed_docs (:obj:`bool`, optional): if :obj:`true`,
            return all SED documents, regardless of whether they have ``master="true"`` or not.
        formats_to_validate (:obj:`list` of :obj:`CombineArchiveContentFormat`, optional): list
            for formats of files to validate
        validate_models_with_languages (:obj:`bool`, optional): if :obj:`True`, validate models
        config (:obj:`Config`, optional): configuration

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors with the archive
            * nested :obj:`list` of :obj:`str`: nested list of warnings with the archive
    """
    if config is None:
        config = get_config()

    errors = []
    warnings = []

    if config.VALIDATE_OMEX_MANIFESTS:
        if not archive.contents:
            errors.append(['Archive must have at least one content element.'])

        # check that locations are listed once per manifest
        locations = set(['manifest.xml'])
        duplicate_locations = set()
        for content in archive.contents:
            if content and content.location:
                location = os.path.relpath(content.location, '.')
                if content.location in locations:
                    duplicate_locations.add(location)
                else:
                    locations.add(location)
        if duplicate_locations:
            errors.append(['The manifest contains repeated content items for these locations:', [
                [location] for location in sorted(duplicate_locations)]])

        # check that all files in the archive are in the manifest
        missing_locations = []
        for dirname, _, filenames in os.walk(archive_dirname):
            for filename in filenames:
                location = os.path.relpath(os.path.join(dirname, filename), archive_dirname)
                if location not in locations:
                    missing_locations.append(location)
        if missing_locations:
            errors.append(['The manifest does not contain content items for these locations:', [
                [location] for location in sorted(missing_locations)]])

        # check for errors with the content elements of the archive
        for i_content, content in enumerate(archive.contents):
            content_errors = []

            if isinstance(content, CombineArchiveContent):
                if content.location:
                    if not os.path.isfile(os.path.join(archive_dirname, content.location)):
                        content_errors.append(['Location is not a file.'])

                else:
                    content_errors.append(['Content element must have a location (e.g., `{}`).'.format(
                        'key-experiment/simulation.sedml')])

                content_errors.extend(validate_format(content.format))

            else:
                content_errors.append(['Contents element must be an instance of `CombineArchiveContent`.'])

            if content_errors:
                content_id = '`' + content.location + '`' if getattr(content, 'location', None) else str(i_content + 1)
                errors.append(['Content element {} is invalid.'.format(content_id), content_errors])

        # check if the archive has at least one SED-ML file that should be executed
        sedml_contents = get_sedml_contents(archive,
                                            include_all_sed_docs_when_no_sed_doc_is_master=include_all_sed_docs_when_no_sed_doc_is_master,
                                            always_include_all_sed_docs=always_include_all_sed_docs)
        if config.VALIDATE_OMEX_MANIFESTS and not sedml_contents:
            warnings.append(['The archive does not contain any SED-ML files that should be executed.'])

    # validate files
    for content in archive.contents:
        if isinstance(content, CombineArchiveContent) and content.location and content.format:
            content_errors, content_warnings = validate_content(
                content, archive_dirname,
                formats_to_validate=formats_to_validate,
                validate_models_with_languages=validate_models_with_languages,
                config=config)
            errors.extend(content_errors)
            warnings.extend(content_warnings)

    # validate OMEX metadata files
    if config.VALIDATE_OMEX_METADATA and CombineArchiveContentFormat.OMEX_METADATA in formats_to_validate:
        _, content_errors, content_warnings = read_omex_meta_files_for_archive(archive, archive_dirname, config=config)
        errors.extend(content_errors)
        warnings.extend(content_warnings)

    # return errors and warnings
    return (errors, warnings)


def validate_format(format):
    """ Validate a COMBINE of an element of a COMBINE/OMEX archive

    Args:
        format (:obj:`str`): format

    Returns:
        nested :obj:`list` of :obj:`str`: nested list of errors with the archive
    """
    errors = []

    if format:
        if not (
            re.match(r'^http://purl\.org/NET/mediatypes/[a-z0-9_\-\+\.]+/[a-z0-9_\-\+\.;]+$', format, re.IGNORECASE)
            or re.match(r'^http://identifiers\.org/combine\.specifications/\w+(\-|\.|\w)*$', format, re.IGNORECASE)
        ):
            errors.append([(
                'The format `{}` of the content element is invalid. '
                'The format must be a persistent URL for an internet media type '
                '(e.g., http://purl.org/NET/mediatypes/text/plain) or for the specifications '
                'of a COMBINE format (e.g., http://identifiers.org/combine.specifications/sbml).'
            ).format(format)])
    else:
        errors.append(['Content element must have a format (e.g., `{}`).'.format(
            CombineArchiveContentFormat.SED_ML)])

    return errors


def validate_content(content, archive_dirname,
                     formats_to_validate=[
                         CombineArchiveContentFormat.SED_ML,
                     ],
                     validate_models_with_languages=True,
                     config=None):
    """ Validate an item of a COMBINE/OMEX archive

    Args:
        content (:obj:`CombineArchiveContent`): item of a COMBINE/OMEX archive
        archive_dirname (:obj:`str`): directory with the content of the archive
        formats_to_validate (:obj:`list` of :obj:`CombineArchiveContentFormat`, optional): list
            for formats of files to validate
        validate_models_with_languages (:obj:`bool`, optional): if :obj:`True`, validate models
        config (:obj:`Config`, optional): configuration

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors with the archive
            * nested :obj:`list` of :obj:`str`: nested list of warnings with the archive
    """
    if config is None:
        config = get_config()

    filename = os.path.join(archive_dirname, content.location)
    file_type = None

    errors = []
    warnings = []

    # validate SED-ML files and models referenced by SED-ML files
    if config.VALIDATE_SEDML:
        if (
            CombineArchiveContentFormat.SED_ML in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.SED_ML.value, content.format)
        ):
            file_type = 'SED-ML'
            reader = SedmlSimulationReader()
            try:
                reader.run(filename, validate_models_with_languages=validate_models_with_languages and config.VALIDATE_SEDML_MODELS)
            except Exception:
                if not reader.errors:
                    raise

            errors = reader.errors or []
            warnings = reader.warnings or []

    # validate images
    if config.VALIDATE_IMAGES:
        if (
            CombineArchiveContentFormat.BMP in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.BMP.value, content.format)
        ):
            file_type = CombineArchiveContentFormat.BMP.name
            if not (os.path.isfile(filename) and imghdr.what(filename) == 'bmp'):
                errors.append(['`{}` is not a valid BMP image.'.format(content.location)])

        elif (
            CombineArchiveContentFormat.GIF in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.GIF.value, content.format)
        ):
            file_type = CombineArchiveContentFormat.GIF.name
            if not (os.path.isfile(filename) and imghdr.what(filename) == 'gif'):
                errors.append(['`{}` is not a valid GIF image.'.format(content.location)])

        elif (
            CombineArchiveContentFormat.JPEG in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.JPEG.value, content.format)
        ):
            file_type = CombineArchiveContentFormat.JPEG.name
            if not (os.path.isfile(filename) and imghdr.what(filename) == 'jpeg'):
                errors.append(['`{}` is not a valid JPEG image.'.format(content.location)])

        elif (
            CombineArchiveContentFormat.PNG in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.PNG.value, content.format)
        ):
            file_type = CombineArchiveContentFormat.PNG.name
            if not (os.path.isfile(filename) and imghdr.what(filename) == 'png'):
                errors.append(['`{}` is not a valid PNG image.'.format(content.location)])

        elif (
            CombineArchiveContentFormat.TIFF in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.TIFF.value, content.format)
        ):
            file_type = CombineArchiveContentFormat.TIFF.name
            if not (os.path.isfile(filename) and imghdr.what(filename) == 'tiff'):
                errors.append(['`{}` is not a valid TIFF image.'.format(content.location)])

        elif (
            CombineArchiveContentFormat.WEBP in formats_to_validate
            and content.format
            and re.match(CombineArchiveContentFormatPattern.WEBP.value, content.format)
        ):
            file_type = CombineArchiveContentFormat.WEBP.name
            if not (os.path.isfile(filename) and imghdr.what(filename) == 'webp'):
                errors.append(['`{}` is not a valid WEBP image.'.format(content.location)])

    if errors:
        errors = [[
            'The {} file at location `{}` is invalid.'.format(file_type, content.location),
            errors,
        ]]
    if warnings:
        warnings = [[
            'The {} file at location `{}` may be invalid.'.format(file_type, content.location),
            warnings,
        ]]

    return (errors, warnings)



""" Utilities for reading and writing COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import (CombineArchiveBase, CombineArchive, CombineArchiveContent,  # noqa: F401
                         CombineArchiveContentFormat, CombineArchiveContentFormatPattern)
from ..archive.io import ArchiveReader
from ..config import get_config, Config  # noqa: F401
from ..data_model import Person
from ..utils.core import flatten_nested_list_of_strings
from ..warnings import warn, BioSimulatorsWarning
import dateutil.parser
import dateutil.tz
import libcombine
import os
import re
import zipfile

__all__ = [
    'CombineArchiveWriter',
    'CombineArchiveReader',
    'CombineArchiveZipReader',
]


class CombineArchiveWriter(object):
    """ Writer for COMBINE/OMEX archives

    Attributes:
        errors (nested :obj:`list` of :obj:`str`): errors
        warnings (nested :obj:`list` of :obj:`str`): warnings
    """

    def __init__(self):
        self.errors = []
        self.warnings = []

    def run(self, archive, in_dir, out_file):
        """ Write an archive to a file

        Args:
            archive (:obj:`CombineArchive`): description of archive
            in_dir (:obj:`str`): directory which contains the files in the archive
            out_file (:obj:`str`): path to save archive

        Raises:
            :obj:`AssertionError`: if files could not be added to the archive or the archive could not be
                saved
        """
        self.errors = []
        self.warnings = []

        # instantiate archive
        archive_comb = libcombine.CombineArchive()

        # set metadata about archive
        self._write_metadata(archive, archive_comb, '.')

        # add files to archive
        for i_content, content in enumerate(archive.contents):
            if not archive_comb.addFile(
                os.path.join(in_dir, content.location),
                content.location,
                content.format if content.format else '',
                content.master
            ):
                content_id = '`' + content.location + '`' if content.location else str(i_content + 1)
                msg = 'Content element {} could not be added to the archive.'.format(content_id)
                self.errors.append([msg])
                raise Exception(msg)
            self._write_metadata(content, archive_comb, content.location)

        # save archive to a file
        if not archive_comb.writeToFile(out_file):
            msg = 'Archive could not be saved.'
            self.errors.append([msg])
            raise Exception(msg)

        errors, warnings = get_combine_errors_warnings(archive_comb.getManifest())
        self.errors.extend(errors)
        self.warnings.extend(warnings)

        if self.warnings:
            warn('COMBINE/OMEX archive may be invalid.\n  ' + flatten_nested_list_of_strings(self.warnings).replace('\n', '\n  '),
                 BioSimulatorsWarning)
        if self.errors:
            raise ValueError('COMBINE/OMEX archive is invalid.\n  ' + flatten_nested_list_of_strings(self.errors).replace('\n', '\n  '))

    def write_manifest(self, contents, filename):
        """ Write an OMEX manifest file

        Args:
            contents (:obj:`list` of :obj:`CombineArchiveContent`): contents of a COMBINE/OMEX archive
            filename (:obj:`str`): path to OMEX manifest file
        """
        manifest = libcombine.CaOmexManifest()
        for content in contents:
            content_comb = manifest.createContent()

            if content.location is not None:
                content_comb.setLocation(content.location)

            if content.format is not None:
                content_comb.setFormat(content.format)

            if content.master is not None:
                content_comb.setMaster(content.master)

        errors, warnings = get_combine_errors_warnings(manifest)
        if warnings:
            msg = 'COMBINE/OMEX archive may be invalid.\n  ' + flatten_nested_list_of_strings(warnings).replace('\n', '\n  ')
            warn(msg, BioSimulatorsWarning)
        if errors:
            msg = 'COMBINE/OMEX archive is invalid.\n  ' + flatten_nested_list_of_strings(errors).replace('\n', '\n  ')
            raise ValueError(msg)

        libcombine.writeOMEXToFile(manifest, filename)

    def _write_metadata(self, obj, archive_comb, filename):
        """ Write metadata about an archive or a file in an archive

        Args:
            obj (:obj:`CombineArchiveBase`): archive or file in an archive
            archive_comb (:obj:`libcombine.CombineArchive`): archive
            filename (:obj:`str`): path of object with archive

        Raises:
            :obj:`NotImplementedError`: of the updated date is not defined
        """
        desc_comb = libcombine.OmexDescription()
        desc_comb.setAbout(filename)
        set_attributes = set()

        if obj.description:
            set_attributes.add('description')
            desc_comb.setDescription(obj.description)

        for author in obj.authors:
            set_attributes.add('authors')
            creator_comb = libcombine.VCard()
            if author.given_name:
                creator_comb.setGivenName(author.given_name)
            if author.family_name:
                creator_comb.setFamilyName(author.family_name)
            desc_comb.addCreator(creator_comb)

        if obj.created:
            set_attributes.add('created')
            date_comb = libcombine.Date()
            date_comb.setDateAsString(obj.created.strftime('%Y-%m-%dT%H:%M:%SZ'))
            desc_comb.setCreated(date_comb)

        if obj.updated:
            set_attributes.add('updated')
            date_comb = libcombine.Date()
            date_comb.setDateAsString(obj.updated.strftime('%Y-%m-%dT%H:%M:%SZ'))
            desc_comb.getModified().append(date_comb)

        if set_attributes:
            if desc_comb.isEmpty():
                msg = 'libCOMBINE requires additional metadata. Only the following attributes are set:\n  {}'.format(
                    '\n  '.join(sorted(set_attributes)))
                self.errors.append([msg])
                raise ValueError(msg)

            archive_comb.addMetadata(filename, desc_comb)


class CombineArchiveReader(object):
    """ Reader for COMBINE/OMEX archives

    Attributes:
        errors (nested :obj:`list` of :obj:`str`): errors
        warnings (nested :obj:`list` of :obj:`str`): warnings
    """

    NONE_DATETIME = '2000-01-01T00:00:00Z'

    def __init__(self):
        self.errors = []
        self.warnings = []

    def run(self, in_file, out_dir, include_omex_metadata_files=True, config=None):
        """ Read an archive from a file

        Args:
            in_file (:obj:`str`): path to archive
            out_dir (:obj:`str`): directory where the contents of the archive should be unpacked
            include_omex_metadata_files (:obj:`bool`, optional): whether to include the OMEX metadata
                file as part of the contents of the archive
            config (:obj:`Config`, optional): configuration

        Returns:
            :obj:`CombineArchive`: description of archive

        Raises:
            :obj:`ValueError`: archive is invalid
        """
        if config is None:
            config = get_config()

        self.errors = []
        self.warnings = []

        if not os.path.isfile(in_file):
            msg = "`{}` is not a file.".format(in_file)
            self.errors.append([msg])
            raise ValueError(msg)

        archive_comb = libcombine.CombineArchive()
        archive_initialized = archive_comb.initializeFromArchive(in_file)
        if archive_initialized:
            errors, warnings = get_combine_errors_warnings(archive_comb.getManifest())
            if config.VALIDATE_OMEX_MANIFESTS:
                self.errors.extend(errors)
                self.warnings.extend(warnings)

        if not archive_initialized or errors:
            if not config.VALIDATE_OMEX_MANIFESTS:
                try:
                    archive = CombineArchiveZipReader().run(in_file, out_dir)
                    msg = '`{}` is a plain zip archive, not a COMBINE/OMEX archive.'.format(in_file)
                    self.warnings.append([msg])
                    warn(msg, BioSimulatorsWarning)
                    return archive
                except ValueError:
                    msg = "`{}` is not a valid COMBINE/OMEX archive.".format(in_file)
                    self.errors.append([msg])
                    raise ValueError(msg)
            elif not self.errors:
                msg = "`{}` is not a valid COMBINE/OMEX archive.".format(in_file)
                self.errors.append([msg])
                raise ValueError(msg)

        # instantiate archive
        archive = CombineArchive()

        # read metadata
        self._read_metadata(archive_comb, '.', archive)

        # read files
        for location in archive_comb.getAllLocations():
            location = location.c_str()
            file_comb = archive_comb.getEntryByLocation(location)

            if file_comb.isSetFormat():
                format = file_comb.getFormat()
            else:
                format = None

            content = CombineArchiveContent(
                location=location,
                format=format,
                master=file_comb.isSetMaster() and file_comb.getMaster(),
            )
            self._read_metadata(archive_comb, location, content)
            archive.contents.append(content)

        # extract files
        archive_comb.extractTo(out_dir)

        # read metadata files skipped by libCOMBINE
        content_locations = set(os.path.relpath(content.location, '.') for content in archive.contents)
        if include_omex_metadata_files:
            for manifest_content in self.read_manifest(os.path.join(out_dir, 'manifest.xml'), in_file, config=config):
                if (
                    manifest_content.format
                    and re.match(CombineArchiveContentFormatPattern.OMEX_METADATA.value, manifest_content.format)
                    and os.path.relpath(manifest_content.location, '.') not in content_locations
                ):
                    archive.contents.append(manifest_content)

        # raise warnings and errors
        if self.warnings:
            warn('COMBINE/OMEX archive may be invalid.\n  ' + flatten_nested_list_of_strings(self.warnings).replace('\n', '\n  '),
                 BioSimulatorsWarning)
        if self.errors:
            raise ValueError('`{}` is not a valid COMBINE/OMEX archive.\n  {}'.format(
                in_file, flatten_nested_list_of_strings(self.errors).replace('\n', '\n  ')))

        # return information about archive
        return archive

    def read_manifest(self, filename, archive_filename=None, config=None):
        """ Read the contents of an OMEX manifest file

        Args:
            filename (:obj:`str`): path to OMEX manifest file
            archive_filename (:obj:`str`, option): path to COMBINE archive
            config (:obj:`Config`, optional): configuration

        Returns:
            :obj:`list` of :obj:`CombineArchiveContent`: contents of the OMEX manifest file
        """
        if config is None:
            config = get_config()

        manifest_comb = libcombine.readOMEXFromFile(filename)
        if not isinstance(manifest_comb, libcombine.CaOmexManifest):
            if config.VALIDATE_OMEX_MANIFESTS or archive_filename is None:
                self.errors(['`{}` could not be read as an OMEX manifest file.'].format(filename))
                return []
            else:
                try:
                    return CombineArchiveZipReader().run(archive_filename).contents
                except ValueError:
                    msg = "`{}` is not a valid zip file.".format(archive_filename)
                    self.errors.append([msg])
                    return []

        errors, warnings = get_combine_errors_warnings(manifest_comb)
        if config.VALIDATE_OMEX_MANIFESTS or archive_filename is None:
            self.errors.extend(errors)
            self.warnings.extend(warnings)
            if errors:
                return []
        else:
            if errors:
                try:
                    return CombineArchiveZipReader().run(archive_filename).contents
                except ValueError:
                    msg = "`{}` is not a valid zip file.".format(archive_filename)
                    self.errors.append([msg])
                    return []

        contents = []
        for content_comb in manifest_comb.getListOfContents():
            content = CombineArchiveContent()

            if content_comb.isSetLocation():
                content.location = content_comb.getLocation()

            if content_comb.isSetFormat():
                content.format = content_comb.getFormat()

            if content_comb.isSetMaster():
                content.master = content_comb.getMaster()

            contents.append(content)
        return contents

    def _read_metadata(self, archive_comb, filename, obj):
        """ Read metadata about an archive or a file in an archive

        Args:
            archive_comb (:obj:`libcombine.CombineArchive`): archive
            filename (:obj:`str`): path to object within archive
            obj (:obj:`CombineArchiveBase`): object to add metadata to
        """
        desc_comb = archive_comb.getMetadataForLocation(filename)
        if not desc_comb.isEmpty():
            obj.description = desc_comb.getDescription() or None

            for creator_comb in desc_comb.getCreators():
                obj.authors.append(Person(
                    given_name=creator_comb.getGivenName() or None,
                    family_name=creator_comb.getFamilyName() or None,
                ))

            created_comb = desc_comb.getCreated().getDateAsString()
            if created_comb == self.NONE_DATETIME:
                obj.created = None
            else:
                obj.created = dateutil.parser.parse(created_comb)

            obj.updated = None
            for modified_comb in desc_comb.getModified():
                updated = dateutil.parser.parse(modified_comb.getDateAsString())
                if obj.updated:
                    obj.updated = max(obj.updated, updated)
                else:
                    obj.updated = updated


class CombineArchiveZipReader(object):
    """ Create a COMBINE/OMEX archive object from a plain zip archive. Set the format of files with
    the extension ``.sedml`` to :obj:`CombineArchiveContentFormat.SED_ML.value`.
    """
    @classmethod
    def run(cls, in_file, out_dir=None):
        """ Read an archive from a zip file

        Args:
            in_file (:obj:`str`): path to archive
            out_dir (:obj:`str`, optional): directory where the contents of the archive should be unpacked

        Returns:
            :obj:`CombineArchive`: description of the archive

        Raises:
            :obj:`ValueError`: archive is invalid
        """
        try:
            zip_archive = ArchiveReader().run(in_file, out_dir)
        except (FileNotFoundError, zipfile.BadZipFile):
            msg = '`{}` is not a valid zip archive.'.format(in_file)
            raise ValueError(msg)

        combine_archive = CombineArchive()
        for file in zip_archive.files:
            combine_archive.contents.append(
                CombineArchiveContent(
                    location=file.archive_path,
                    format=(CombineArchiveContentFormat.SED_ML.value
                            if os.path.splitext(file.archive_path)[1] == '.sedml' else None),
                ),
            )

        return combine_archive


def get_combine_errors_warnings(manifest):
    """ Get the errors and warnings of an OMEX manifest

    Args:
        manifest (:obj:`libcombine.CaOmexManifest`): manifest

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: errors
            * nested :obj:`list` of :obj:`str`: warnings
    """
    errors = []
    warnings = []

    log = manifest.getErrorLog()
    for i_error in range(log.getNumErrors()):
        error = log.getError(i_error)
        if error.isError() or error.isFatal():
            errors.append([error.getMessage()])
        else:
            warnings.append([error.getMessage()])

    return (errors, warnings)



""" Utilities for validating NeuroML models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-05-10
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...log.data_model import StandardOutputErrorCapturerLevel  # noqa: E402
from ...log.utils import StandardOutputErrorCapturer  # noqa: E402
from neuroml.loaders import NeuroMLLoader
import pyneuroml.pynml


__all__ = ['validate_model', 'validate_model_l1', 'validate_model_l2']


def validate_model(filename, name=None):
    """ Check that a model is valid

    Args:
        filename (:obj:`str`): path to model
        name (:obj:`str`, optional): name of model for use in error messages

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * :obj:`neuroml.nml.nml.NeuroMLDocument`: model
    """
    l1_valid, l1_output = validate_model_l1(filename)
    if l1_valid:
        return ([], [], NeuroMLLoader.load(filename))

    l2_valid, l2_output = validate_model_l2(filename)
    if l2_valid:
        return ([], [], NeuroMLLoader.load(filename))

    return ([[l2_output]], [], None)


def validate_model_l1(filename):
    """ Check that a file is a valid NeuroML L1 model

    Args:
        filename (:obj:`str`): path to model

    Returns:
        :obj:`tuple`:

            * :obj:`bool`: whether the file is valid
            * :obj:`str`: error message
    """
    with StandardOutputErrorCapturer(level=StandardOutputErrorCapturerLevel.python, relay=False) as captured:
        valid = pyneuroml.pynml.validate_neuroml1(filename)
        output = captured.get_text()
    return (valid, output)


def validate_model_l2(filename):
    """ Check that a file is a valid NeuroML L2 model

    Args:
        filename (:obj:`str`): path to model

    Returns:
        :obj:`tuple`:

            * :obj:`bool`: whether the file is valid
            * :obj:`str`: error message
    """
    with StandardOutputErrorCapturer(level=StandardOutputErrorCapturerLevel.python, relay=False) as captured:
        valid = pyneuroml.pynml.validate_neuroml2(filename)
        output = captured.get_text()
    return (valid, output)



""" Utilities for reading and writing reports

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
from ..sedml.data_model import Output, Report, Plot2D, Plot3D  # noqa: F401
from ..utils.core import pad_arrays_to_consistent_shapes
from ..warnings import warn
from .data_model import DataSetResults, ReportFormat
from .warnings import (RepeatDataSetLabelsWarning, MissingReportMetadataWarning, MissingDataWarning,
                       ExtraDataWarning, CannotExportMultidimensionalTableWarning)
import enum
import functools
import glob
import h5py
import numpy
import openpyxl
import os
import pandas

__all__ = [
    'ReportWriter',
    'ReportReader',
]


class Hdf5DataSetType(enum.Enum):
    """ Type of data encoded in an HDF5 data set """
    SedReport = Report
    SedPlot2D = Plot2D
    SedPlot3D = Plot3D


class ReportWriter(object):
    """ Class for writing reports of simulation results """

    def run(self, report, results, base_path, rel_path, format=ReportFormat.h5, type=Report):
        """ Save a report

        Args:
            report (:obj:`Report`): report
            results (:obj:`DataSetResults`): results of the data sets
            base_path (:obj:`str`): path to save results

                * CSV: parent directory to save results
                * HDF5: file to save results

            rel_path (:obj:`str`): path to save results relative to :obj:`base_path`

                * CSV: relative path to :obj:`base_path`
                * HDF5: key within HDF5 file

            format (:obj:`ReportFormat`, optional): report format
            type (:obj:`type`): type of output (e.g., subclass of :obj:`Output` such as :obj:`Report`, :obj:`Plot2D`)
        """
        rel_path = os.path.relpath(rel_path, '.')

        results_array = []
        data_set_ids = []
        data_set_labels = []
        data_set_names = []
        data_set_data_types = []
        data_set_shapes = []
        for data_set in report.data_sets:
            if data_set.id in results:
                data_set_result = results[data_set.id]
                results_array.append(data_set_result)
                data_set_ids.append(data_set.id)
                data_set_labels.append(data_set.label)
                data_set_names.append(data_set.name or '')
                if data_set_result is None:
                    data_set_data_types.append('__None__')
                    data_set_shapes.append('')
                else:
                    data_set_dtype = data_set_result.dtype
                    if data_set_dtype in [numpy.dtype('object'), numpy.dtype('void'), numpy.dtype('S'), numpy.dtype('a')]:
                        msg = 'NumPy dtype should be a specific type such as `float64` or `int64` not `{}`.'.format(data_set_dtype.name)
                        raise TypeError(msg)
                    data_set_data_types.append(data_set_dtype.name)
                    data_set_shapes.append(','.join(str(dim_len) for dim_len in data_set_result.shape))
        results_array = pad_arrays_to_consistent_shapes(results_array)
        results_array = numpy.array(results_array)

        if format in [ReportFormat.csv, ReportFormat.tsv, ReportFormat.xlsx]:
            if results_array.ndim > 2:
                msg = 'Report has {} dimensions. Multidimensional reports cannot be exported to {}.'.format(
                    results_array.ndim, format.value.upper())
                warn(msg, CannotExportMultidimensionalTableWarning)
                return

            if len(set(data_set.label for data_set in report.data_sets)) < len(report.data_sets):
                warn('To facilitate machine interpretation, data sets should have unique labels.',
                     RepeatDataSetLabelsWarning)

            msg = 'Reports exported to {} do not contain information about the data type or size of each data set.'.format(
                format.value.upper())
            warn(msg, MissingReportMetadataWarning)

            results_df = pandas.DataFrame(results_array, index=data_set_labels)

            if format in [ReportFormat.csv, ReportFormat.tsv]:
                filename = os.path.join(base_path, rel_path + '.' + format.value)
                out_dir = os.path.dirname(filename)
                if not os.path.isdir(out_dir):
                    os.makedirs(out_dir)

                results_df.to_csv(filename, header=False, sep=',' if format == ReportFormat.csv else '\t')
            else:
                filename = os.path.join(base_path, os.path.dirname(rel_path) + '.' + format.value)
                out_dir = os.path.dirname(filename)
                if not os.path.isdir(out_dir):
                    os.makedirs(out_dir)

                with pandas.ExcelWriter(filename, mode='a' if os.path.isfile(filename) else 'w', engine='openpyxl') as writer:
                    results_df.to_excel(writer, sheet_name=os.path.basename(rel_path), header=False)

        elif format == ReportFormat.h5:
            filename = os.path.join(base_path, get_config().H5_REPORTS_PATH)
            if not os.path.isdir(base_path):
                os.makedirs(base_path)

            rel_path = '/'.join(rel_path.split(os.path.sep))

            with h5py.File(filename, 'a') as file:
                try:
                    file[rel_path]
                    del file[rel_path]
                except KeyError:
                    pass

                data_set = file.create_dataset(rel_path, data=results_array,
                                               chunks=True, compression="gzip", compression_opts=9)
                data_set.attrs['_type'] = Hdf5DataSetType(type).name
                if report.id:
                    data_set.attrs['uri'] = rel_path
                    data_set.attrs['sedmlId'] = report.id
                if report.name:
                    data_set.attrs['sedmlName'] = report.name
                data_set.attrs['sedmlDataSetIds'] = data_set_ids
                data_set.attrs['sedmlDataSetNames'] = data_set_names
                data_set.attrs['sedmlDataSetLabels'] = data_set_labels
                data_set.attrs['sedmlDataSetDataTypes'] = data_set_data_types
                data_set.attrs['sedmlDataSetShapes'] = data_set_shapes

                group_ids = rel_path.split('/')[0:-1]
                for i_group in range(len(group_ids)):
                    uri = '/'.join(group_ids[0:i_group + 1])
                    group = file[uri]
                    group.attrs['uri'] = uri
                    group.attrs['combineArchiveLocation'] = uri

        else:
            raise NotImplementedError('Report format {} is not supported'.format(format))


class ReportReader(object):
    """ Class for reading reports of simulation results """

    def run(self, report, base_path, rel_path, format=ReportFormat.h5):
        """ Read a report for a file

        Args:
            report (:obj:`Report`): report
            base_path (:obj:`str`): path to save results

                * CSV: parent directory to save results
                * HDF5: file to save results

            rel_path (:obj:`str`): path to save results relative to :obj:`base_path`

                * CSV: relative path to :obj:`base_path`
                * HDF5: key within HDF5 file

            format (:obj:`ReportFormat`, optional): report format

        Returns:
            :obj:`DataSetResults`: report results
        """
        rel_path = os.path.relpath(rel_path, '.')

        if format in [ReportFormat.csv, ReportFormat.tsv, ReportFormat.xlsx]:
            warn('Reports exported to {} do not contain information about the data type or size of each data set.'.format(
                format.value.upper()), MissingReportMetadataWarning)

            if format in [ReportFormat.csv, ReportFormat.tsv]:
                filename = os.path.join(base_path, rel_path + '.' + format.value)
                df = pandas.read_csv(filename,
                                     index_col=0,
                                     header=None,
                                     sep=',' if format == ReportFormat.csv else '\t')
            else:
                filename = os.path.join(base_path, os.path.dirname(rel_path) + '.' + format.value)
                df = pandas.read_excel(filename,
                                       sheet_name=os.path.basename(rel_path),
                                       index_col=0,
                                       header=None,
                                       engine='openpyxl')
            df.columns = pandas.RangeIndex(start=0, stop=df.shape[1], step=1)

            results = DataSetResults()

            data_set_labels = [data_set.label for data_set in report.data_sets]
            if df.index.tolist() == data_set_labels:
                data = df.to_numpy()
                for i_data_set, data_set in enumerate(report.data_sets):
                    results[data_set.id] = data[i_data_set, :]
                extra_data_sets = set()

            else:
                data_set_label_to_index = {}
                for i_data_set, data_set_label in enumerate(df.index):
                    if data_set_label not in data_set_label_to_index:
                        data_set_label_to_index[data_set_label] = i_data_set
                    else:
                        data_set_label_to_index[data_set_label] = None

                unreadable_data_sets = []
                for data_set in report.data_sets:
                    i_data_set = data_set_label_to_index.get(data_set.label, None)
                    if i_data_set is None:
                        # results[data_set.id] = None
                        unreadable_data_sets.append(data_set.id)
                    else:
                        results[data_set.id] = df.loc[data_set.label, :].to_numpy()

                if unreadable_data_sets:
                    warn('Some data sets could not be read because their labels are not unique:\n  - {}'.format(
                        '\n'.join('`' + id + '`' for id in sorted(unreadable_data_sets))), RepeatDataSetLabelsWarning)

                data_set_id_to_label = {data_set.id: data_set.label for data_set in report.data_sets}
                extra_data_sets = set(df.index) - set(data_set_id_to_label[id] for id in results.keys()) - set(unreadable_data_sets)

            file_data_set_ids = set(results.keys()) | extra_data_sets

        elif format == ReportFormat.h5:
            filename = os.path.join(base_path, get_config().H5_REPORTS_PATH)

            rel_path = '/'.join(rel_path.split(os.path.sep))

            with h5py.File(filename, 'r') as file:
                data_set = file[rel_path]
                data_set_results = data_set[:]
                file_data_set_ids = data_set.attrs['sedmlDataSetIds']
                data_set_data_types = data_set.attrs['sedmlDataSetDataTypes']
                data_set_shapes = []
                for data_set_shape in data_set.attrs['sedmlDataSetShapes']:
                    if data_set_shape:
                        data_set_shapes.append([int(dim_len) for dim_len in data_set_shape.split(',')])
                    else:
                        data_set_shapes.append([])

            results = DataSetResults()
            data_set_id_to_index = {data_set_id: i_data_set for i_data_set, data_set_id in enumerate(file_data_set_ids)}

            data_set_ndim = data_set_results.ndim - 1
            for data_set in report.data_sets:
                i_data_set = data_set_id_to_index.get(data_set.id, None)
                if i_data_set is not None:
                    data_set_data_type = data_set_data_types[i_data_set]
                    if data_set_data_type == '__None__':
                        results[data_set.id] = None
                    else:
                        data_set_shape = data_set_shapes[i_data_set]
                        data_set_slice = [slice(0, dim_len) for dim_len in data_set_shape] + \
                            [slice(0, 1)] * (data_set_ndim - len(data_set_shape))
                        results[data_set.id] = (
                            data_set_results[i_data_set][data_set_slice]
                            .reshape(data_set_shape)
                            .astype(data_set_data_type)
                        )

            file_data_set_ids = set(file_data_set_ids)

        else:
            raise NotImplementedError('Report format {} is not supported'.format(format))

        report_data_set_ids = set(data_set.id for data_set in report.data_sets)
        missing_data_set_ids = report_data_set_ids.difference(file_data_set_ids)
        extra_data_set_ids = file_data_set_ids.difference(report_data_set_ids)

        if missing_data_set_ids:
            warn('File does not contain data for the following data sets of the report:\n  - {}'.format(
                '\n'.join('`' + id + '`' for id in sorted(missing_data_set_ids))), MissingDataWarning)

        if extra_data_set_ids:
            warn('File contains additional data that could not be mapped to data sets of the report:\n  - {}'.format(
                '\n'.join('`' + id + '`' for id in sorted(extra_data_set_ids))), ExtraDataWarning)

        return results

    def get_ids(self, base_path, format=ReportFormat.h5, type=Output):
        """ Get the ids of the reports in a file

        Args:
            base_path (:obj:`str`): path to save results

                * CSV: parent directory to save results
                * HDF5: file to save results

            format (:obj:`ReportFormat`, optional): report format
            type (:obj:`type`): type of report to get

        Returns:
            :obj:`list` of :obj:`str`: ids of reports
        """
        if format in [ReportFormat.csv, ReportFormat.tsv]:
            report_ids = []
            for path in glob.glob(os.path.join(base_path, '**/*.' + format.value), recursive=True):
                report_ids.append(os.path.relpath(path, base_path)[0:-len(format.value)-1])
            return report_ids

        elif format == ReportFormat.xlsx:
            report_ids = []
            for path in glob.glob(os.path.join(base_path, '**/*.' + format.value), recursive=True):
                wb = openpyxl.load_workbook(path)
                for sheet_name in wb.get_sheet_names():
                    report_ids.append(os.path.join(os.path.relpath(path, base_path)[0:-len(format.value)-1], sheet_name))
            return report_ids

        elif format == ReportFormat.h5:
            filename = os.path.join(base_path, get_config().H5_REPORTS_PATH)
            with h5py.File(filename, 'r') as file:
                report_ids = []
                file.visititems(functools.partial(self.append_report_id, type, report_ids))

            return report_ids

        else:
            raise NotImplementedError('Report format {} is not supported'.format(format))

    @staticmethod
    def append_report_id(type, report_ids, name, object):
        if isinstance(object, h5py.Dataset):
            data_set_type = object.attrs.get('_type', None)
            if (
                data_set_type and
                (
                    Hdf5DataSetType[data_set_type].value == type
                    or issubclass(Hdf5DataSetType[data_set_type].value, type)
                )
            ):
                report_ids.append(name)



""" Exceptions for SED-ML

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-12
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..exceptions import BioSimulatorsException

__all__ = [
    'SedmlExecutionError',
    'UnsupportedModelLanguageError',
]


class SedmlExecutionError(BioSimulatorsException):
    """ Error that a SED document could not be executed """
    pass  # pragma: no cover


class UnsupportedModelLanguageError(BioSimulatorsException, NotImplementedError):
    """ Error that a SED document could not be executed """
    pass  # pragma: no cover



""" Warnings for plots

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-20
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..warnings import BioSimulatorsWarning

__all__ = [
    'IllogicalVizWarning',
]


class IllogicalVizWarning(BioSimulatorsWarning):
    """ Warning that a plot is may be illogical (e.g., has mixed axes or scales) """
    pass  # pragma: no cover



""" Common configuration for simulation tools

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .omex_meta.data_model import OmexMetadataInputFormat, OmexMetadataOutputFormat, OmexMetadataSchema
from .report.data_model import ReportFormat  # noqa: F401
from .viz.data_model import VizFormat  # noqa: F401
from kisao import AlgorithmSubstitutionPolicy  # noqa: F401
import appdirs
import enum
import os

__all__ = ['Config', 'get_config', 'Colors', 'get_app_dirs']

DEFAULT_OMEX_METADATA_INPUT_FORMAT = OmexMetadataInputFormat.rdfxml
DEFAULT_OMEX_METADATA_OUTPUT_FORMAT = OmexMetadataOutputFormat.rdfxml_abbrev
DEFAULT_OMEX_METADATA_SCHEMA = OmexMetadataSchema.biosimulations
DEFAULT_ALGORITHM_SUBSTITUTION_POLICY = AlgorithmSubstitutionPolicy.SIMILAR_VARIABLES
DEFAULT_H5_REPORTS_PATH = 'reports.h5'
DEFAULT_REPORTS_PATH = 'reports.zip'
DEFAULT_PLOTS_PATH = 'plots.zip'
DEFAULT_LOG_PATH = 'log.yml'
DEFAULT_BIOSIMULATORS_API_ENDPOINT = 'https://api.biosimulators.org/'
DEFAULT_RUNBIOSIMULATIONS_API_ENDPOINT = 'https://api.biosimulations.org/'


class Config(object):
    """ Configuration

    Attributes:
        OMEX_METADATA_INPUT_FORMAT (:obj:`OmexMetadataInputFormat`): format to validate OMEX Metadata files against
        OMEX_METADATA_OUTPUT_FORMAT (:obj:`OmexMetadataOutputFormat`): format to export OMEX Metadata files
        OMEX_METADATA_SCHEMA (:obj:`OmexMetadataSchema`): schema to validate OMEX Metadata files against
        VALIDATE_OMEX_MANIFESTS (:obj:`bool`): whether to validate OMEX manifests during the validation of COMBINE/OMEX archives
        VALIDATE_SEDML (:obj:`bool`): whether to validate SED-ML files during the validation of COMBINE/OMEX archives
        VALIDATE_SEDML_MODELS (:obj:`bool`): whether to validate models referenced by SED-ML files during the validation of COMBINE/OMEX archives
        VALIDATE_OMEX_METADATA (:obj:`bool`): whether to validate OMEX metadata (RDF files) during the validation of COMBINE/OMEX archives
        VALIDATE_IMAGES (:obj:`bool`): whether to validate the images in COMBINE/OMEX archives during their validation
        VALIDATE_RESULTS (:obj:`bool`): whether to validate the results of simulations following their execution
        ALGORITHM_SUBSTITUTION_POLICY (:obj:`AlgorithmSubstitutionPolicy`): algorithm substition policy
        COLLECT_COMBINE_ARCHIVE_RESULTS (:obj:`bool`): whether to assemble an in memory data structure with all of the simulation results
            of COMBINE/OMEX archives
        COLLECT_SED_DOCUMENT_RESULTS (:obj:`bool`): whether to assemble an in memory data structure with all of the simulation results
            of SED documents
        SAVE_PLOT_DATA (:obj:`bool`): whether to save data for plots alongside data for reports in CSV/HDF5 files
        REPORT_FORMATS (:obj:`list` of :obj:`ReportFormat`): default formats to generate reports in
        VIZ_FORMATS (:obj:`list` of :obj:`VizFormat`): default formats to generate plots in
        H5_REPORTS_PATH (:obj:`str`): path to save reports in HDF5 format relative to base output directory
        REPORTS_PATH (:obj:`str`): path to save zip archive of reports relative to base output directory
        PLOTS_PATH (:obj:`str`): path to save zip archive of plots relative to base output directory
        BUNDLE_OUTPUTS (:obj:`bool`): indicates whether bundles of report and plot outputs should be produced
        KEEP_INDIVIDUAL_OUTPUTS (:obj:`bool`): indicates whether the individual output files should be kept
        LOG (:obj:`bool`): whether to log the execution of a COMBINE/OMEX archive
        LOG_PATH (:obj:`str`): path to save the execution log of a COMBINE/OMEX archive
        BIOSIMULATORS_API_ENDPOINT (:obj:`str`): URL for BioSimulators API
        RUNBIOSIMULATIONS_API_ENDPOINT (:obj:`str`): URL for runBioSimulations API
        VERBOSE (:obj:`bool`): whether to display the detailed output of the execution of each task
        DEBUG (:obj:`bool`): whether to raise exceptions rather than capturing them
    """

    def __init__(self,
                 OMEX_METADATA_INPUT_FORMAT=DEFAULT_OMEX_METADATA_INPUT_FORMAT,
                 OMEX_METADATA_OUTPUT_FORMAT=DEFAULT_OMEX_METADATA_OUTPUT_FORMAT,
                 OMEX_METADATA_SCHEMA=DEFAULT_OMEX_METADATA_SCHEMA,
                 VALIDATE_OMEX_MANIFESTS=True,
                 VALIDATE_SEDML=True,
                 VALIDATE_SEDML_MODELS=True,
                 VALIDATE_OMEX_METADATA=True,
                 VALIDATE_IMAGES=True,
                 VALIDATE_RESULTS=True,
                 ALGORITHM_SUBSTITUTION_POLICY=DEFAULT_ALGORITHM_SUBSTITUTION_POLICY,
                 COLLECT_COMBINE_ARCHIVE_RESULTS=False,
                 COLLECT_SED_DOCUMENT_RESULTS=False,
                 SAVE_PLOT_DATA=True,
                 REPORT_FORMATS=[ReportFormat.h5],
                 VIZ_FORMATS=[VizFormat.pdf],
                 H5_REPORTS_PATH=DEFAULT_H5_REPORTS_PATH,
                 REPORTS_PATH=DEFAULT_REPORTS_PATH,
                 PLOTS_PATH=DEFAULT_PLOTS_PATH,
                 BUNDLE_OUTPUTS=True,
                 KEEP_INDIVIDUAL_OUTPUTS=True,
                 LOG=True,
                 LOG_PATH=DEFAULT_LOG_PATH,
                 BIOSIMULATORS_API_ENDPOINT=DEFAULT_BIOSIMULATORS_API_ENDPOINT,
                 RUNBIOSIMULATIONS_API_ENDPOINT=DEFAULT_RUNBIOSIMULATIONS_API_ENDPOINT,
                 VERBOSE=False,
                 DEBUG=False):
        """
        Args:
            OMEX_METADATA_INPUT_FORMAT (:obj:`OmexMetadataInputFormat`, optional): format to validate OMEX Metadata files against
            OMEX_METADATA_OUTPUT_FORMAT (:obj:`OmexMetadataOutputFormat`, optional): format to export OMEX Metadata files
            OMEX_METADATA_SCHEMA (:obj:`OmexMetadataSchema`, optional): schema to validate OMEX Metadata files against
            VALIDATE_OMEX_MANIFESTS (:obj:`bool`, optional): whether to validate OMEX manifests during the execution of COMBINE/OMEX archives
            VALIDATE_SEDML (:obj:`bool`, optional): whether to validate SED-ML files during the execution of COMBINE/OMEX archives
            VALIDATE_SEDML_MODELS (:obj:`bool`, optional): whether to validate models referenced by SED-ML files during the execution
                of COMBINE/OMEX archives
            VALIDATE_OMEX_METADATA (:obj:`bool`, optional): whether to validate OMEX metadata (RDF files) during the execution of
                COMBINE/OMEX archives
            VALIDATE_IMAGES (:obj:`bool`, optional): whether to validate the images in COMBINE/OMEX archives during their execution
            VALIDATE_RESULTS (:obj:`bool`, optional): whether to validate the results of simulations following their execution
            ALGORITHM_SUBSTITUTION_POLICY (:obj:`str`, optional): algorithm substition policy
            COLLECT_COMBINE_ARCHIVE_RESULTS (:obj:`bool`, optional): whether to assemble an in memory data structure with all of the
                simulation results of COMBINE/OMEX archives
            COLLECT_SED_DOCUMENT_RESULTS (:obj:`bool`, optional): whether to assemble an in memory data structure with all of the
                simulation results of SED documents
            SAVE_PLOT_DATA (:obj:`bool`, optional): whether to save data for plots alongside data for reports in CSV/HDF5 files
            REPORT_FORMATS (:obj:`list` of :obj:`str`, optional): default formats to generate reports in
            VIZ_FORMATS (:obj:`list` of :obj:`str`, optional): default formats to generate plots in
            H5_REPORTS_PATH (:obj:`str`, optional): path to save reports in HDF5 format relative to base output directory
            REPORTS_PATH (:obj:`str`, optional): path to save zip archive of reports relative to base output directory
            PLOTS_PATH (:obj:`str`, optional): path to save zip archive of plots relative to base output directory
            BUNDLE_OUTPUTS (:obj:`bool`, optional): indicates whether bundles of report and plot outputs should be produced
            KEEP_INDIVIDUAL_OUTPUTS (:obj:`bool`, optional): indicates whether the individual output files should be kept
            LOG (:obj:`bool`, optional): whether to log the execution of a COMBINE/OMEX archive
            LOG_PATH (:obj:`str`, optional): path to save the execution status of a COMBINE/OMEX archive
            BIOSIMULATORS_API_ENDPOINT (:obj:`str`, optional): URL for BioSimulators API
            RUNBIOSIMULATIONS_API_ENDPOINT (:obj:`str`, optional): URL for runBioSimulations API
            VERBOSE (:obj:`bool`, optional): whether to display the detailed output of the execution of each task
            DEBUG (:obj:`bool`, optional): whether to raise exceptions rather than capturing them
        """
        self.OMEX_METADATA_INPUT_FORMAT = OMEX_METADATA_INPUT_FORMAT
        self.OMEX_METADATA_OUTPUT_FORMAT = OMEX_METADATA_OUTPUT_FORMAT
        self.OMEX_METADATA_SCHEMA = OMEX_METADATA_SCHEMA
        self.VALIDATE_OMEX_MANIFESTS = VALIDATE_OMEX_MANIFESTS
        self.VALIDATE_SEDML = VALIDATE_SEDML
        self.VALIDATE_SEDML_MODELS = VALIDATE_SEDML_MODELS
        self.VALIDATE_OMEX_METADATA = VALIDATE_OMEX_METADATA
        self.VALIDATE_IMAGES = VALIDATE_IMAGES
        self.VALIDATE_RESULTS = VALIDATE_RESULTS
        self.ALGORITHM_SUBSTITUTION_POLICY = ALGORITHM_SUBSTITUTION_POLICY
        self.COLLECT_COMBINE_ARCHIVE_RESULTS = COLLECT_COMBINE_ARCHIVE_RESULTS
        self.COLLECT_SED_DOCUMENT_RESULTS = COLLECT_SED_DOCUMENT_RESULTS
        self.SAVE_PLOT_DATA = SAVE_PLOT_DATA
        self.REPORT_FORMATS = REPORT_FORMATS
        self.VIZ_FORMATS = VIZ_FORMATS
        self.H5_REPORTS_PATH = H5_REPORTS_PATH
        self.REPORTS_PATH = REPORTS_PATH
        self.PLOTS_PATH = PLOTS_PATH
        self.BUNDLE_OUTPUTS = BUNDLE_OUTPUTS
        self.KEEP_INDIVIDUAL_OUTPUTS = KEEP_INDIVIDUAL_OUTPUTS
        self.LOG = LOG
        self.LOG_PATH = LOG_PATH
        self.BIOSIMULATORS_API_ENDPOINT = BIOSIMULATORS_API_ENDPOINT
        self.RUNBIOSIMULATIONS_API_ENDPOINT = RUNBIOSIMULATIONS_API_ENDPOINT
        self.VERBOSE = VERBOSE
        self.DEBUG = DEBUG


def get_config():
    """ Get the configuration

    Returns:
        :obj:`Config`: configuration
    """
    report_formats = os.environ.get('REPORT_FORMATS', 'h5').strip()
    if report_formats:
        report_formats = [ReportFormat(format.strip().lower()) for format in report_formats.split(',')]
    else:
        report_formats = []

    viz_formats = os.environ.get('VIZ_FORMATS', 'pdf').strip()
    if viz_formats:
        viz_formats = [VizFormat(format.strip().lower()) for format in viz_formats.split(',')]
    else:
        viz_formats = []

    return Config(
        OMEX_METADATA_INPUT_FORMAT=OmexMetadataInputFormat(os.environ.get(
            'OMEX_METADATA_INPUT_FORMAT', DEFAULT_OMEX_METADATA_INPUT_FORMAT)),
        OMEX_METADATA_OUTPUT_FORMAT=OmexMetadataOutputFormat(os.environ.get(
            'OMEX_METADATA_OUTPUT_FORMAT', DEFAULT_OMEX_METADATA_OUTPUT_FORMAT)),
        OMEX_METADATA_SCHEMA=OmexMetadataSchema(os.environ.get(
            'OMEX_METADATA_SCHEMA', DEFAULT_OMEX_METADATA_SCHEMA)),
        VALIDATE_OMEX_MANIFESTS=os.environ.get('VALIDATE_OMEX_MANIFESTS', '1').lower() in ['1', 'true'],
        VALIDATE_SEDML=os.environ.get('VALIDATE_SEDML', '1').lower() in ['1', 'true'],
        VALIDATE_SEDML_MODELS=os.environ.get('VALIDATE_SEDML_MODELS', '1').lower() in ['1', 'true'],
        VALIDATE_OMEX_METADATA=os.environ.get('VALIDATE_OMEX_METADATA', '1').lower() in ['1', 'true'],
        VALIDATE_IMAGES=os.environ.get('VALIDATE_IMAGES', '1').lower() in ['1', 'true'],
        VALIDATE_RESULTS=os.environ.get('VALIDATE_RESULTS', '1').lower() in ['1', 'true'],
        ALGORITHM_SUBSTITUTION_POLICY=AlgorithmSubstitutionPolicy(os.environ.get(
            'ALGORITHM_SUBSTITUTION_POLICY', DEFAULT_ALGORITHM_SUBSTITUTION_POLICY)),
        COLLECT_COMBINE_ARCHIVE_RESULTS=os.environ.get('COLLECT_COMBINE_ARCHIVE_RESULTS', '0').lower() in ['1', 'true'],
        COLLECT_SED_DOCUMENT_RESULTS=os.environ.get('COLLECT_SED_DOCUMENT_RESULTS', '0').lower() in ['1', 'true'],
        SAVE_PLOT_DATA=os.environ.get('SAVE_PLOT_DATA', '1').lower() in ['1', 'true'],
        REPORT_FORMATS=report_formats,
        VIZ_FORMATS=viz_formats,
        H5_REPORTS_PATH=os.environ.get('H5_REPORTS_PATH', DEFAULT_H5_REPORTS_PATH),
        REPORTS_PATH=os.environ.get('REPORTS_PATH', DEFAULT_REPORTS_PATH),
        PLOTS_PATH=os.environ.get('PLOTS_PATH', DEFAULT_PLOTS_PATH),
        BUNDLE_OUTPUTS=os.environ.get('BUNDLE_OUTPUTS', '1').lower() in ['1', 'true'],
        KEEP_INDIVIDUAL_OUTPUTS=os.environ.get('KEEP_INDIVIDUAL_OUTPUTS', '1').lower() in ['1', 'true'],
        LOG=os.environ.get('LOG', '1').lower() in ['1', 'true'],
        LOG_PATH=os.environ.get('LOG_PATH', DEFAULT_LOG_PATH),
        BIOSIMULATORS_API_ENDPOINT=os.environ.get('BIOSIMULATORS_API_ENDPOINT', DEFAULT_BIOSIMULATORS_API_ENDPOINT),
        RUNBIOSIMULATIONS_API_ENDPOINT=os.environ.get('RUNBIOSIMULATIONS_API_ENDPOINT', DEFAULT_RUNBIOSIMULATIONS_API_ENDPOINT),
        VERBOSE=os.environ.get('VERBOSE', '1').lower() in ['1', 'true'],
        DEBUG=os.environ.get('DEBUG', '0').lower() in ['1', 'true'],
    )


Colors = enum.Enum('Colors',
                   {
                       'queued': 'cyan',
                       'success': 'blue',
                       'succeeded': 'blue',
                       'running': 'green',
                       'pass': 'green',
                       'passed': 'green',
                       'failure': 'red',
                       'failed': 'red',
                       'skip': 'magenta',
                       'skipped': 'magenta',
                       'warning': 'yellow',
                       'warned': 'yellow',
                   },
                   type=str)


def get_app_dirs():
    """ Get the directories for the application

    Returns:
        :obj:`appdirs.AppDirs`: application directories
    """
    return appdirs.AppDirs("BioSimulatorsUtils", "BioSimulatorsTeam")



""" Utilities for working with Smoldyn models

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-04-05
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ...sedml.data_model import (SedDocument, ModelAttributeChange, Variable,  # noqa: F401
                                 Symbol, Simulation, UniformTimeCourseSimulation, Algorithm,
                                 Task)
from ...utils.core import flatten_nested_list_of_strings
from .validation import validate_model
from smoldyn.biosimulators.utils import read_simulation
import os
import re
import types  # noqa: F401

__all__ = ['get_parameters_variables_outputs_for_simulation']


def get_parameters_variables_outputs_for_simulation(model_filename, model_language, simulation_type, algorithm_kisao_id=None,
                                                    change_level=SedDocument, native_ids=False, native_data_types=False):
    """ Get the possible observables for a simulation of a model

    Args:
        model_filename (:obj:`str`): path to model file
        model_language (:obj:`str`): model language (e.g., ``urn:sedml:language:sbml``)
        simulation_type (:obj:`types.Type`): subclass of :obj:`Simulation`
        algorithm_kisao_id (:obj:`str`): KiSAO id of the algorithm for simulating the model (e.g., ``KISAO_0000019``
            for CVODE)
        change_level (:obj:`types.Type`, optional): level at which model changes will be made (:obj:`SedDocument` or :obj:`Task`)
        native_ids (:obj:`bool`, optional): whether to return the raw id and name of each model component rather than the suggested name
            for the variable of an associated SED-ML data generator
        native_data_types (:obj:`bool`, optional): whether to return new_values in their native data types

    Returns:
        :obj:`list` of :obj:`ModelAttributeChange`: possible attributes of a model that can be changed and their default values
        :obj:`list` of :obj:`Simulation`: simulation of the model
        :obj:`list` of :obj:`Variable`: possible observables for a simulation of the model
    """
    # check model file exists and is valid
    if not isinstance(model_filename, str):
        raise ValueError('`{}` is not a path to a model file.'.format(model_filename))

    if not os.path.isfile(model_filename):
        raise FileNotFoundError('Model file `{}` does not exist.'.format(model_filename))

    errors, _, (smoldyn_model, model_config) = validate_model(model_filename)
    if errors:
        raise ValueError('Model file `{}` is not a valid Smoldyn file.\n  {}'.format(
            model_filename, flatten_nested_list_of_strings(errors).replace('\n', '\n  ')))

    if simulation_type not in [UniformTimeCourseSimulation]:
        raise NotImplementedError('`simulation_type` must be `OneStepSimulation` or `UniformTimeCourseSimulation`')

    sim = UniformTimeCourseSimulation(
        id='simulation',
        initial_time=smoldyn_model.start,
        output_start_time=smoldyn_model.start,
        output_end_time=smoldyn_model.stop,
        number_of_steps=int((smoldyn_model.stop - smoldyn_model.start) / smoldyn_model.dt),
        algorithm=Algorithm(
            kisao_id=algorithm_kisao_id or 'KISAO_0000057',
        ),
    )

    # get parameters and observables
    model = read_simulation(model_filename)

    params = []
    for instruction in model.instructions:
        if not (
            instruction.macro.startswith('define ')
            # or instruction.macro.startswith('define_global ')
            or instruction.macro.startswith('difc ')
            or instruction.macro.startswith('difc_rule ')
            or instruction.macro.startswith('difm ')
            or instruction.macro.startswith('difm_rule ')
            or instruction.macro.startswith('drift ')
            or instruction.macro.startswith('drift_rule ')
            or instruction.macro.startswith('surface_drift ')
            or instruction.macro.startswith('surface_drift_rule ')
        ):
            continue

        if native_data_types:
            id = instruction.macro.partition(' ')[2]
        else:
            id = re.sub(r'[^a-zA-Z0-9_]', '_', instruction.macro)

        if native_data_types:
            if (
                instruction.macro.startswith('define ')
                # or instruction.macro.startswith('define_global ')
                or instruction.macro.startswith('difc ')
                or instruction.macro.startswith('difc_rule ')
            ):
                new_value = float(instruction.arguments)
            elif (
                instruction.macro.startswith('difm ')
                or instruction.macro.startswith('difm_rule ')
                or instruction.macro.startswith('drift ')
                or instruction.macro.startswith('drift_rule ')
                or instruction.macro.startswith('surface_drift ')
                or instruction.macro.startswith('surface_drift_rule ')
            ):
                new_value = [float(val) for val in instruction.arguments.split(' ')]
        else:
            new_value = instruction.arguments

        if instruction.macro.partition(' ')[2] != 'all':
            params.append(ModelAttributeChange(
                id=id,
                name=None if native_ids else instruction.description,
                target=instruction.macro,
                new_value=new_value,
            ))

    smoldyn_model.addOutputData('counts')
    smoldyn_model.addCommand(cmd='molcount counts', cmd_type='E')
    for compartment in model.compartments:
        data_id = 'counts_cmpt_' + compartment
        smoldyn_model.addOutputData(data_id)
        smoldyn_model.addCommand(cmd='molcountincmpt ' + compartment + ' ' + data_id, cmd_type='E')
    for surface in model.surfaces:
        data_id = 'counts_surf_' + surface
        smoldyn_model.addOutputData(data_id)
        smoldyn_model.addCommand(cmd='molcountonsurf ' + surface + ' ' + data_id, cmd_type='E')

    smoldyn_model.run(stop=1e-12, dt=1., overwrite=True, display=False, quit_at_end=False)

    data_id = 'counts'
    species_counts = smoldyn_model.getOutputData(data_id, True)[0][1:]
    for species, count in zip(model.species, species_counts):
        params.append(ModelAttributeChange(
            id=species if native_ids else 'initial_count_species_{}'.format(re.sub('[^a-zA-Z0-9_]', '_', species)),
            name=None if native_ids else 'Initial count of species "{}"'.format(species),
            target="fixmolcount {}".format(species),
            new_value=count if native_data_types else str(count),
        ))

    for compartment in model.compartments:
        data_id = 'counts_cmpt_' + compartment
        species_counts = smoldyn_model.getOutputData(data_id, True)[0][1:]
        for species, count in zip(model.species, species_counts):
            params.append(ModelAttributeChange(
                id="{}.{}".format(species, compartment) if native_ids else 'initial_count_species_{}_compartment_{}'.format(
                    re.sub('[^a-zA-Z0-9_]', '_', species), compartment),
                name=None if native_ids else 'Initial count of species "{}" in compartment "{}"'.format(species, compartment),
                target="fixmolcountincmpt {} {}".format(species, compartment),
                new_value=new_value,
            ))

    for surface in model.surfaces:
        data_id = 'counts_surf_' + surface
        species_counts = smoldyn_model.getOutputData(data_id, True)[0][1:]
        for species, count in zip(model.species, species_counts):
            params.append(ModelAttributeChange(
                id="{}.{}".format(species, surface) if native_ids else 'initial_count_species_{}_surface_{}'.format(
                    re.sub('[^a-zA-Z0-9_]', '_', species), surface),
                name=None if native_ids else 'Initial count of species "{}" in surface "{}"'.format(species, surface),
                target="fixmolcountonsurf {} {}".format(species, surface),
                new_value=new_value,
            ))

    vars = []
    vars.append(Variable(
        id=None if native_ids else 'time',
        name=None if native_ids else 'Time',
        symbol=Symbol.time.value,
    ))
    for species in model.species:
        vars.append(Variable(
            id=species if native_ids else 'count_species_{}'.format(re.sub('[^a-zA-Z0-9_]', '_', species)),
            name=None if native_ids else 'Count of species "{}"'.format(species),
            target="molcount {}".format(species),
        ))
        for compartment in model.compartments:
            vars.append(Variable(
                id="{}.{}".format(species, compartment) if native_ids else 'count_species_{}_compartment_{}'.format(
                    re.sub('[^a-zA-Z0-9_]', '_', species), compartment),
                name=None if native_ids else 'Count of species "{}" in compartment "{}"'.format(species, compartment),
                target="molcountincmpt {} {}".format(species, compartment),
            ))
        for surface in model.surfaces:
            vars.append(Variable(
                id="{}.{}".format(species, surface) if native_ids else 'count_species_{}_surface_{}'.format(
                    re.sub('[^a-zA-Z0-9_]', '_', species), surface),
                name=None if native_ids else 'Count of species "{}" in surface "{}"'.format(species, surface),
                target="molcountonsurf {} {}".format(species, surface),
            ))

    return (params, [sim], vars, [])



""" Utilities for executing tasks in SED documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config, Config, Colors  # noqa: F401
from ..log.data_model import Status, SedDocumentLog, TaskLog, ReportLog, Plot2DLog, Plot3DLog, StandardOutputErrorCapturerLevel  # noqa: F401
from ..log.utils import init_sed_document_log, StandardOutputErrorCapturer
from ..report.data_model import VariableResults, DataSetResults, ReportResults, ReportFormat  # noqa: F401
from ..report.io import ReportWriter
from ..utils.core import pad_arrays_to_consistent_shapes
from ..viz.data_model import VizFormat  # noqa: F401
from ..viz.io import write_plot_2d, write_plot_3d
from ..warnings import warn
from .data_model import SedDocument, Model, Task, RepeatedTask, Output, Report, Plot2D, Plot3D, ModelAttributeChange, DataSet  # noqa: F401
from .exceptions import SedmlExecutionError
from .io import SedmlSimulationReader
from .utils import (resolve_model_and_apply_xml_changes, get_variables_for_task,
                    calc_data_generators_results, resolve_range, get_models_referenced_by_task,
                    get_value_of_variable_model_xml_targets, calc_compute_model_change_new_value,
                    apply_changes_to_xml_model, get_first_last_models_executed_by_task,
                    is_model_language_encoded_in_xml)
from .warnings import NoTasksWarning, NoOutputsWarning, SedmlFeatureNotSupportedWarning
from lxml import etree  # noqa: F401
import copy
import datetime
import numpy
import os
import sys
import tempfile
import termcolor
import types  # noqa: F401


__all__ = [
    'exec_sed_doc',
    'exec_task',
    'exec_repeated_task',
    'exec_report',
    'exec_plot_2d',
    'exec_plot_3d',
    'get_report_for_plot2d',
    'get_report_for_plot3d',
]


def exec_sed_doc(task_executer, doc, working_dir, base_out_path, rel_out_path=None,
                 apply_xml_model_changes=False,
                 log=None, indent=0, pretty_print_modified_xml_models=False,
                 log_level=StandardOutputErrorCapturerLevel.c,
                 config=None):
    """ Execute the tasks specified in a SED document and generate the specified outputs

    Args:
        task_executer (:obj:`types.FunctionType`): function to execute each task in the SED-ML file.
            The function must implement the following interface::

                def exec_task(task, variables, preprocessed_task=None, log=None, config=None, **simulator_config):
                    ''' Execute a simulation and return its results

                    Args:
                        task (:obj:`Task`): task
                        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
                        preprocessed_task (:obj:`object`, optional): preprocessed information about the task, including possible
                            model changes and variables. This can be used to avoid repeatedly executing the same initialization
                            for repeated calls to this method.
                        log (:obj:`TaskLog`, optional): log for the task
                        config (:obj:`Config`, optional): BioSimulators common configuration
                        **simulator_config (:obj:`dict`, optional): optional simulator-specific configuration

                    Returns:
                       :obj:`VariableResults`: results of variables
                    '''
                    pass

        doc (:obj:`SedDocument` or :obj:`str`): SED document or a path to SED-ML file which defines a SED document
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)

        base_out_path (:obj:`str`): path to store the outputs

            * CSV: directory in which to save outputs to files
              ``{base_out_path}/{rel_out_path}/{report.id}.csv``
            * HDF5: directory in which to save a single HDF5 file (``{base_out_path}/reports.h5``),
              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file

        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the outputs
        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
            calling :obj:`task_executer`.
        log (:obj:`SedDocumentLog`, optional): log of the document
        indent (:obj:`int`, optional): degree to indent status messages
        pretty_print_modified_xml_models (:obj:`bool`, optional): if :obj:`True`, pretty print modified XML models
        log_level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which to log output
        config (:obj:`Config`): configuration

    Returns:
        :obj:`tuple`:

            * :obj:`ReportResults`: results of each report
            * :obj:`SedDocumentLog`: log of the document
    """
    if not config:
        config = get_config()

    # process arguments
    if not isinstance(doc, SedDocument):
        doc = SedmlSimulationReader().run(doc)
    else:
        doc = copy.deepcopy(doc)

    if config.LOG and not log:
        log = init_sed_document_log(doc)

    verbose = config.VERBOSE

    # update status
    exceptions = []

    # execute tasks
    if not doc.tasks:
        warn('SED document does not describe any tasks.', NoTasksWarning)

    # TODO: initialize reports with their eventual shapes; this requires individual simulation tools to pass
    # information about the shape of their output to this method
    variable_results = VariableResults()

    if config.COLLECT_SED_DOCUMENT_RESULTS:
        report_results = ReportResults()
    else:
        report_results = None

    print('{}Found {} tasks and {} outputs:\n{}Tasks:\n{}{}\n{}Outputs:\n{}{}'.format(
        ' ' * 2 * indent,
        len(doc.tasks),
        len(doc.outputs),
        ' ' * 2 * (indent + 1),
        ' ' * 2 * (indent + 2),
        ('\n' + ' ' * 2 * (indent + 2)).join(sorted('`' + task.id + '`' for task in doc.tasks)),
        ' ' * 2 * (indent + 1),
        ' ' * 2 * (indent + 2),
        ('\n' + ' ' * 2 * (indent + 2)).join(sorted('`' + output.id + '`' for output in doc.outputs)),
    ))
    for i_task, task in enumerate(doc.tasks):
        print('{}Executing task {}: `{}`'.format(' ' * 2 * indent, i_task + 1, task.id))

        if config.LOG:
            task_log = log.tasks[task.id]
            task_log.status = Status.RUNNING
            task_log.export()
        else:
            task_log = None

        # Execute task
        print('{}Executing simulation ...'.format(' ' * 2 * (indent + 1)), end='')
        sys.stdout.flush()
        with StandardOutputErrorCapturer(relay=verbose, level=log_level, disabled=not config.LOG) as captured:
            start_time = datetime.datetime.now()
            try:
                # get model and apply changes
                original_models = get_models_referenced_by_task(task)
                original_model_sources = {}
                original_model_changes = {}
                temp_model_sources = []
                model_etrees = {}
                for original_model in original_models:
                    original_model_sources[original_model.id] = original_model.source
                    original_model_changes[original_model.id] = original_model.changes

                    temp_model, temp_model_source, model_etree = resolve_model_and_apply_xml_changes(
                        original_model, doc, working_dir,
                        apply_xml_model_changes=apply_xml_model_changes,
                        pretty_print_modified_xml_models=pretty_print_modified_xml_models)

                    original_model.source = temp_model.source
                    original_model.changes = temp_model.changes

                    if temp_model_source:
                        temp_model_sources.append(temp_model_source)

                    model_etrees[original_model.id] = model_etree

                task_vars = get_variables_for_task(doc, task)

                # execute task
                if isinstance(task, Task):
                    task_var_results = exec_task(task, task_executer, task_vars, doc, log=task_log, config=config)

                elif isinstance(task, RepeatedTask):
                    task_var_results = exec_repeated_task(task, task_executer, task_vars, doc,
                                                          apply_xml_model_changes=apply_xml_model_changes,
                                                          model_etrees=model_etrees,
                                                          pretty_print_modified_xml_models=pretty_print_modified_xml_models,
                                                          config=config)

                else:  # pragma: no cover: already validated by :obj:`get_models_referenced_by_task`
                    raise NotImplementedError('Tasks of type {} are not supported.'.format(task.__class__.__name__))

                # append results
                for key, value in task_var_results.items():
                    variable_results[key] = value

                # log status
                task_status = Status.SUCCEEDED
                task_exception = None

                # cleanup modified model sources
                for temp_model_source in temp_model_sources:
                    os.remove(temp_model_source)
                for original_model in original_models:
                    original_model.source = original_model_sources[original_model.id]
                    original_model.changes = original_model_changes[original_model.id]
            except Exception as exception:
                if config.DEBUG:
                    raise
                exceptions.append(exception)
                task_status = Status.FAILED
                task_exception = exception

        if config.LOG:
            task_log.status = task_status
            task_log.exception = task_exception
            task_log.output = captured.get_text()
            task_log.duration = (datetime.datetime.now() - start_time).total_seconds()
            task_log.export()
        print(' ' + termcolor.colored(task_status.value.lower(), Colors[task_status.value.lower()].value))

        # generate outputs
        print('{}Generating {} outputs ...'.format(' ' * 2 * (indent + 1), len(doc.outputs)))
        task_contributes_to_output = False
        report_formats = config.REPORT_FORMATS
        viz_formats = config.VIZ_FORMATS
        for i_output, output in enumerate(doc.outputs):
            print('{}Generating output {}: `{}` ...'.format(' ' * 2 * (indent + 2), i_output + 1, output.id), end='')
            sys.stdout.flush()
            start_time = datetime.datetime.now()
            with StandardOutputErrorCapturer(relay=verbose, level=log_level, disabled=not config.LOG) as captured:
                try:
                    if config.LOG and log.outputs[output.id].status == Status.SUCCEEDED:
                        output_status = log.outputs[output.id].status
                        print(' ' + termcolor.colored(output_status.value.lower(), Colors[output_status.value.lower()].value))
                        continue

                    if isinstance(output, Report):
                        output_result, output_status, output_exception, task_contributes_to_report = exec_report(
                            output, variable_results,
                            base_out_path, rel_out_path, report_formats,
                            task=task,
                            log=log.outputs[output.id] if config.LOG else None,
                            type=Report)
                        task_contributes_to_output = task_contributes_to_output or task_contributes_to_report

                    elif isinstance(output, Plot2D):
                        output_status, output_exception, task_contributes_to_plot = exec_plot_2d(
                            output, variable_results,
                            base_out_path, rel_out_path, viz_formats,
                            task=task,
                            log=log.outputs[output.id] if config.LOG else None)
                        task_contributes_to_output = task_contributes_to_output or task_contributes_to_plot

                        # save data as report
                        if config.SAVE_PLOT_DATA:
                            report = get_report_for_plot2d(output)
                            output_result, _, _, _ = exec_report(
                                report, variable_results,
                                base_out_path, rel_out_path, report_formats,
                                task,
                                log=None,
                                type=output.__class__)
                        else:
                            output_result = None

                    elif isinstance(output, Plot3D):
                        output_status, output_exception, task_contributes_to_plot = exec_plot_3d(
                            output, variable_results,
                            base_out_path, rel_out_path, viz_formats,
                            task=task,
                            log=log.outputs[output.id] if config.LOG else None)
                        task_contributes_to_output = task_contributes_to_output or task_contributes_to_plot

                        # save as report
                        if config.SAVE_PLOT_DATA:
                            report = get_report_for_plot3d(output)
                            output_result, _, _, _ = exec_report(
                                report, variable_results,
                                base_out_path, rel_out_path, report_formats,
                                task,
                                log=None,
                                type=output.__class__)
                        else:
                            output_result = None

                    else:
                        # unreachable because the above cases cover all types of outputs
                        raise NotImplementedError('Outputs of type {} are not supported.'.format(output.__class__.__name__))

                    if config.COLLECT_SED_DOCUMENT_RESULTS and output_result is not None:
                        report_results[output.id] = output_result

                except Exception as exception:
                    if config.DEBUG:
                        raise
                    output_status = Status.FAILED
                    output_exception = exception

            if config.LOG:
                log.outputs[output.id].status = output_status
                log.outputs[output.id].exception = output_exception
                log.outputs[output.id].output = captured.get_text()
                log.outputs[output.id].duration = (datetime.datetime.now() - start_time).total_seconds()
                log.outputs[output.id].export()

            if output_exception:
                exceptions.append(output_exception)

            print(' ' + termcolor.colored(output_status.value.lower(), Colors[output_status.value.lower()].value))

        if not task_contributes_to_output:
            warn('Task {} does not contribute to any outputs.'.format(task.id), NoOutputsWarning)

    # finalize the status of the outputs
    if config.LOG:
        for output_log in log.outputs.values():
            output_log.finalize()

    # summarize execution
    if config.LOG:
        task_status_count = {
            Status.SUCCEEDED: 0,
            Status.SKIPPED: 0,
            Status.FAILED: 0,
        }
        for task_log in log.tasks.values():
            task_status_count[task_log.status] += 1

        output_status_count = {
            Status.SUCCEEDED: 0,
            Status.SKIPPED: 0,
            Status.FAILED: 0,
        }
        for output_log in log.outputs.values():
            output_status_count[output_log.status] += 1

        print('')
        print('{}Executed {} tasks and {} outputs:'.format(' ' * 2 * indent, len(doc.tasks), len(doc.outputs)))
        print('{}  Tasks:'.format(' ' * 2 * indent))
        print('{}    Succeeded: {}'.format(' ' * 2 * indent, task_status_count[Status.SUCCEEDED]))
        print('{}    Skipped: {}'.format(' ' * 2 * indent, task_status_count[Status.SKIPPED]))
        print('{}    Failed: {}'.format(' ' * 2 * indent, task_status_count[Status.FAILED]))
        print('{}  Outputs:'.format(' ' * 2 * indent))
        print('{}    Succeeded: {}'.format(' ' * 2 * indent, output_status_count[Status.SUCCEEDED]))
        print('{}    Skipped: {}'.format(' ' * 2 * indent, output_status_count[Status.SKIPPED]))
        print('{}    Failed: {}'.format(' ' * 2 * indent, output_status_count[Status.FAILED]))

    # raise exceptions
    if exceptions:
        msg = 'The SED document did not execute successfully:\n\n  {}'.format(
            '\n\n  '.join(str(exceptions).replace('\n', '\n  ') for exceptions in exceptions))
        raise SedmlExecutionError(msg)

    # return the results of the reports
    return report_results, log


def exec_task(task, task_executer, task_vars, doc, log=None, config=None):
    """ Execute a basic SED task

    Args:
        task (:obj:`Task`): task
        task_executer (:obj:`types.FunctionType`): function to execute each task in the SED-ML file.
            The function must implement the following interface::

                def exec_task(task, variables, preprocessed_task=None, log=None, config=None, **simulator_config):
                    ''' Execute a simulation and return its results

                    Args:
                        task (:obj:`Task`): task
                        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
                        preprocessed_task (:obj:`object`, optional): preprocessed information about the task, including possible
                            model changes and variables. This can be used to avoid repeatedly executing the same initialization
                            for repeated calls to this method.
                        log (:obj:`TaskLog`, optional): log for the task
                        config (:obj:`Config`, optional): BioSimulators common configuration
                        **simulator_config (:obj:`dict`, optional): optional simulator-specific configuration

                    Returns:
                        :obj:`tuple`:

                            * :obj:`VariableResults`: results of variables
                            * :obj:`TaskLog`: log
                    '''
                    pass

        task_vars (:obj:`list` of :obj:`Variable`): variables that task must record
        doc (:obj:`SedDocument` or :obj:`str`): SED document or a path to SED-ML file which defines a SED document
        log (:obj:`TaskLog`, optional): log
        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`VariableResults`: results of the variables
    """
    # execute task
    task_variable_results, _ = task_executer(task, task_vars, log=log, config=config)

    # check that the expected variables were recorded
    variable_results = VariableResults()
    for var in task_vars:
        variable_results[var.id] = task_variable_results.get(var.id, None)

    # return results
    return variable_results


def exec_repeated_task(task, task_executer, task_vars, doc, apply_xml_model_changes=False, model_etrees=None,
                       pretty_print_modified_xml_models=False, config=None):
    """ Execute a repeated SED task

    Args:
        task (:obj:`RepeatedTask`): task
        task_executer (:obj:`types.FunctionType`): function to execute each task in the SED-ML file.
            The function must implement the following interface::

                def exec_task(task, variables, preprocessed_task=None, log=None, config=None, **simulator_config):
                    ''' Execute a simulation and return its results

                    Args:
                       task (:obj:`Task`): task
                       variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
                       preprocessed_task (:obj:`object`, optional): preprocessed information about the task, including possible
                            model changes and variables. This can be used to avoid repeatedly executing the same initialization
                            for repeated calls to this method.
                       log (:obj:`TaskLog`, optional): log for the task
                       config (:obj:`Config`, optional): BioSimulators common configuration
                       **simulator_config (:obj:`dict`, optional): optional simulator-specific configuration

                    Returns:
                       :obj:`VariableResults`: results of variables
                    '''
                    pass

        task_vars (:obj:`list` of :obj:`Variable`): variables that task must record
        doc (:obj:`SedDocument` or :obj:`str`): SED document or a path to SED-ML file which defines a SED document
        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
            calling :obj:`task_executer`.
        model_etrees (:obj:`dict` of :obj:`str` to :obj:`etree._Element`)
        pretty_print_modified_xml_models (:obj:`bool`, optional): if :obj:`True`, pretty print modified XML models
        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`VariableResults`: results of the variables
    """
    # warn about inability to not reset models
    if not task.reset_model_for_each_iteration:
        models = get_first_last_models_executed_by_task(task)
        if models[0] == models[-1]:
            msg = (
                'Only independent execution of iterations of repeated tasks is supported. '
                'Successive iterations will not be executed starting from the end state of the previous iteration.'
            )
            warn(msg, SedmlFeatureNotSupportedWarning)

    sub_tasks = sorted(task.sub_tasks, key=lambda sub_task: sub_task.order)
    for prev_sub_task, next_sub_task in zip(sub_tasks[0:-1], sub_tasks[1:]):
        if get_first_last_models_executed_by_task(prev_sub_task.task)[-1] == get_first_last_models_executed_by_task(next_sub_task.task)[0]:
            msg = (
                'Only independent execution of sub-tasks is supported. '
                'Successive sub-tasks will not be executed starting from the end state of the previous sub-task.'
            )
            warn(msg, SedmlFeatureNotSupportedWarning)
            break

    # hold onto model to be able to reset it
    if task.reset_model_for_each_iteration:
        original_doc = doc
        original_task = task
        original_model_etrees = model_etrees

    # resolve the ranges
    main_range_values = resolve_range(task.range, model_etrees=model_etrees)

    range_values = {}
    for range in task.ranges:
        range_values[range.id] = resolve_range(range, model_etrees=model_etrees)
    for change in task.changes:
        if change.range:
            range_values[change.range.id] = resolve_range(change.range, model_etrees=model_etrees)

    # initialize the results of the sub-tasks
    variable_results = VariableResults()
    for var in task_vars:
        variable_results[var.id] = []
        for main_range_value in main_range_values:
            variable_results[var.id].append([None] * len(task.sub_tasks))

    # iterate over the main range, apply the changes to the model(s), execute the sub-tasks, and record the results of the tasks
    for i_main_range, _ in enumerate(main_range_values):
        # reset the models referenced by the task
        if task.reset_model_for_each_iteration:
            doc = copy.deepcopy(original_doc)
            task = next(task for task in doc.tasks if task.id == original_task.id)
            model_etrees = copy.deepcopy(original_model_etrees)

        # get range values
        current_range_values = {}
        current_range_values[task.range.id] = range_values[task.range.id][i_main_range]
        for range in task.ranges:
            current_range_values[range.id] = range_values[range.id][i_main_range]
        for change in task.changes:
            if change.range:
                current_range_values[change.range.id] = range_values[change.range.id][i_main_range]

        # apply the changes to the models
        for change in task.changes:
            variable_values = {}
            for variable in change.variables:
                if not apply_xml_model_changes:
                    raise NotImplementedError('Set value changes that involve variables of non-XML-encoded models are not supported.')
                variable_values[variable.id] = get_value_of_variable_model_xml_targets(variable, model_etrees)

            new_value = calc_compute_model_change_new_value(change, variable_values=variable_values, range_values=current_range_values)
            if new_value == int(new_value):
                new_value = str(int(new_value))
            else:
                new_value = str(new_value)

            if change.symbol:
                raise NotImplementedError('Set value changes of symbols is not supported.')

            attr_change = ModelAttributeChange(target=change.target, target_namespaces=change.target_namespaces, new_value=new_value)

            if apply_xml_model_changes and is_model_language_encoded_in_xml(change.model.language):
                model = Model(changes=[attr_change])
                apply_changes_to_xml_model(model, model_etrees[change.model.id], None, None)

            else:
                change.model.changes.append(attr_change)

        # sort the sub-tasks
        sub_tasks = sorted(task.sub_tasks, key=lambda sub_task: sub_task.order)

        # execute the sub-tasks and record their results
        for i_sub_task, sub_task in enumerate(sub_tasks):
            if isinstance(sub_task.task, Task):
                model = sub_task.task.model
                if apply_xml_model_changes and is_model_language_encoded_in_xml(model.language):
                    original_model_source = model.source
                    fid, model.source = tempfile.mkstemp(suffix='.xml')
                    os.close(fid)

                    model_etrees[model.id].write(model.source,
                                                 xml_declaration=True,
                                                 encoding="utf-8",
                                                 standalone=False,
                                                 pretty_print=pretty_print_modified_xml_models)

                sub_task_var_results = exec_task(sub_task.task, task_executer, task_vars, doc, config=config)

                if apply_xml_model_changes and is_model_language_encoded_in_xml(model.language):
                    os.remove(model.source)
                    model.source = original_model_source

            elif isinstance(sub_task.task, RepeatedTask):
                sub_task_var_results = exec_repeated_task(sub_task.task, task_executer, task_vars, doc,
                                                          apply_xml_model_changes=apply_xml_model_changes,
                                                          model_etrees=model_etrees,
                                                          pretty_print_modified_xml_models=pretty_print_modified_xml_models,
                                                          config=config)

            else:  # pragma: no cover: already validated by :obj:`get_first_last_models_executed_by_task`
                raise NotImplementedError('Tasks of type {} are not supported.'.format(sub_task.task.__class__.__name__))

            for var in task_vars:
                variable_results[var.id][i_main_range][i_sub_task] = sub_task_var_results.get(var.id, None)

    # shape results to consistent size
    arrays = []
    for var in task_vars:
        for i_main_range, _ in enumerate(main_range_values):
            for i_sub_task, sub_task in enumerate(sub_tasks):
                arrays.append(variable_results[var.id][i_main_range][i_sub_task])

    padded_arrays = pad_arrays_to_consistent_shapes(arrays)

    i_array = 0
    for var in task_vars:
        for i_main_range, _ in enumerate(main_range_values):
            for i_sub_task, sub_task in enumerate(sub_tasks):
                variable_results[var.id][i_main_range][i_sub_task] = padded_arrays[i_array]
                i_array += 1

        variable_results[var.id] = numpy.array(variable_results[var.id])

    # return the results of the task
    return variable_results


def exec_report(report, variable_results, base_out_path, rel_out_path, formats, task, log=None, type=Report):
    """ Execute a report, generating the data sets which are available

    Args:
        report (:obj:`Report`): report
        variable_results (:obj:`VariableResults`): result of each data generator
        base_out_path (:obj:`str`): path to store the outputs

            * CSV: directory in which to save outputs to files
              ``{base_out_path}/{rel_out_path}/{report.id}.csv``
            * HDF5: directory in which to save a single HDF5 file (``{base_out_path}/reports.h5``),
              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file

        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the outputs
        formats (:obj:`list` of :obj:`ReportFormat`, optional): report format (e.g., csv or h5)
        task (:obj:`Task`): task
        log (:obj:`ReportLog`, optional): log of report
        type (:obj:`types.Type`): type of output (e.g., subclass of :obj:`Output` such as :obj:`Report`, :obj:`Plot2D`)

    Returns:
        :obj:`tuple`:

            * :obj:`DataSetResults`: report
            * :obj:`Status`: status
            * :obj:`Exception`: exception for failure
            * :obj:`bool`: whether :obj:`task` contribute a variable to the report
    """
    # calculate data generators
    data_generators = set()
    for data_set in report.data_sets:
        data_generators.add(data_set.data_generator)

    data_gen_results, data_gen_statuses, data_gen_exceptions, task_contributes_to_report = calc_data_generators_results(
        data_generators, variable_results, report, task, make_shapes_consistent=False)

    # collect data sets
    data_set_results = {}

    running = False
    succeeded = True
    failed = False

    for data_set in report.data_sets:
        data_gen_res = data_gen_results[data_set.data_generator.id]
        data_set_results[data_set.id] = data_gen_res

        data_gen_status = data_gen_statuses[data_set.data_generator.id]
        if log:
            log.data_sets[data_set.id] = data_gen_status
        if data_gen_status == Status.FAILED:
            failed = True
        if data_gen_status == Status.SUCCEEDED:
            running = True
        else:
            succeeded = False

    for format in formats:
        ReportWriter().run(report,
                           data_set_results,
                           base_out_path,
                           os.path.join(rel_out_path, report.id) if rel_out_path else report.id,
                           format=format,
                           type=type)

    if failed:
        status = Status.FAILED

    elif running:
        if succeeded:
            status = Status.SUCCEEDED
        else:
            status = Status.RUNNING

    else:
        status = Status.QUEUED

    return data_set_results, status, data_gen_exceptions, task_contributes_to_report


def exec_plot_2d(plot, variable_results, base_out_path, rel_out_path, formats, task, log):
    """ Execute a 2D plot, generating the curves which are available

    Args:
        plot (:obj:`Plot2D`): plot
        variable_results (:obj:`VariableResults`): result of each data generator
        base_out_path (:obj:`str`): base path to store the plot. Complete path is
            ``{base_out_path}/{rel_out_path}/{plot.id}.csv``
        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the plot
        formats (:obj:`list` of :obj:`VizFormat`, optional): plot format (e.g., pdf)
        task (:obj:`Task`): task
        log (:obj:`ReportLog`, optional): log of plot

    Returns:
        :obj:`tuple`:

            * :obj:`Status`: status
            * :obj:`Exception`: exception for failure
            * :obj:`bool`: whether :obj:`task` contributes a variable to the plot
    """
    # calculate data generators
    data_generators = set()
    for curve in plot.curves:
        data_generators.add(curve.x_data_generator)
        data_generators.add(curve.y_data_generator)

    data_gen_results, data_gen_statuses, data_gen_exceptions, task_contributes_to_plot = calc_data_generators_results(
        data_generators, variable_results, plot, task)

    # collect data sets
    running = False
    succeeded = True
    failed = False

    succeeded_curves = []
    for curve in plot.curves:
        x_data_gen_status = data_gen_statuses[curve.x_data_generator.id]
        y_data_gen_status = data_gen_statuses[curve.y_data_generator.id]

        if x_data_gen_status == Status.SUCCEEDED and y_data_gen_status == Status.SUCCEEDED:
            curve_status = Status.SUCCEEDED
            succeeded_curves.append(curve)
        elif x_data_gen_status == Status.FAILED or y_data_gen_status == Status.FAILED:
            curve_status = Status.FAILED
        else:
            curve_status = Status.QUEUED

        if log:
            log.curves[curve.id] = curve_status

        if curve_status == Status.FAILED:
            failed = True
        if curve_status == Status.SUCCEEDED:
            running = True
        else:
            succeeded = False

    for format in formats:
        write_plot_2d(Plot2D(curves=succeeded_curves),
                      data_gen_results,
                      base_out_path,
                      os.path.join(rel_out_path, plot.id) if rel_out_path else plot.id,
                      format=format)

    # determine the overall status of the plot
    if failed:
        status = Status.FAILED

    elif running:
        if succeeded:
            status = Status.SUCCEEDED
        else:
            status = Status.RUNNING

    else:
        status = Status.QUEUED

    # return
    return status, data_gen_exceptions, task_contributes_to_plot


def exec_plot_3d(plot, variable_results, base_out_path, rel_out_path, formats, task, log):
    """ Execute a 3D plot, generating the surfaces which are available

    Args:
        plot (:obj:`Plot3D`): plot
        variable_results (:obj:`VariableResults`): result of each data generator
        base_out_path (:obj:`str`): base path to store the plot. Complete path is
          ``{base_out_path}/{rel_out_path}/{plot.id}.pdf``
        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the plot
        formats (:obj:`list` of :obj:`VizFormat`, optional): plot format (e.g., pdf)
        task (:obj:`Task`): task
        log (:obj:`ReportLog`, optional): log of plot

    Returns:
        :obj:`tuple`:

            * :obj:`Status`: status
            * :obj:`Exception`: exception for failure
            * :obj:`bool`: whether :obj:`task` contributes a variable to the plot
    """
    # calculate data generators
    data_generators = set()
    for surface in plot.surfaces:
        data_generators.add(surface.x_data_generator)
        data_generators.add(surface.y_data_generator)
        data_generators.add(surface.z_data_generator)

    data_gen_results, data_gen_statuses, data_gen_exceptions, task_contributes_to_plot = calc_data_generators_results(
        data_generators, variable_results, plot, task)

    # collect data sets
    running = False
    succeeded = True
    failed = False

    succeeded_surfaces = []
    for surface in plot.surfaces:
        x_data_gen_status = data_gen_statuses[surface.x_data_generator.id]
        y_data_gen_status = data_gen_statuses[surface.y_data_generator.id]
        z_data_gen_status = data_gen_statuses[surface.z_data_generator.id]

        if x_data_gen_status == Status.SUCCEEDED and y_data_gen_status == Status.SUCCEEDED and z_data_gen_status == Status.SUCCEEDED:
            surface_status = Status.SUCCEEDED
            succeeded_surfaces.append(surface)
        elif x_data_gen_status == Status.FAILED or y_data_gen_status == Status.FAILED or z_data_gen_status == Status.FAILED:
            surface_status = Status.FAILED
        else:
            surface_status = Status.QUEUED

        if log:
            log.surfaces[surface.id] = surface_status

        if surface_status == Status.FAILED:
            failed = True
        if surface_status == Status.SUCCEEDED:
            running = True
        else:
            succeeded = False

    for format in formats:
        write_plot_3d(Plot3D(surfaces=succeeded_surfaces),
                      data_gen_results,
                      base_out_path,
                      os.path.join(rel_out_path, plot.id) if rel_out_path else plot.id,
                      format=format)

    # determine the overall status of the plot
    if failed:
        status = Status.FAILED

    elif running:
        if succeeded:
            status = Status.SUCCEEDED
        else:
            status = Status.RUNNING

    else:
        status = Status.QUEUED

    # return
    return status, data_gen_exceptions, task_contributes_to_plot


def get_report_for_plot2d(plot):
    """ Get a report for a 2D plot with a dataset for each data generator of the curves

    Args:
        plot (:obj:`Plot2D`): plot

    Returns:
        :obj:`Report`: report with a dataset for each data generator of the curves
    """
    data_sets = {}
    for curve in plot.curves:
        data_sets[curve.x_data_generator.id] = DataSet(
            id=curve.x_data_generator.id,
            name=curve.x_data_generator.name,
            label=curve.x_data_generator.id,
            data_generator=curve.x_data_generator,
        )
        data_sets[curve.y_data_generator.id] = DataSet(
            id=curve.y_data_generator.id,
            name=curve.y_data_generator.name,
            label=curve.y_data_generator.id,
            data_generator=curve.y_data_generator,
        )
    return Report(
        id=plot.id,
        name=plot.name,
        data_sets=sorted(data_sets.values(), key=lambda data_set: data_set.id))


def get_report_for_plot3d(plot):
    """ Get a report for a 3D plot with a dataset for each data generator of the surfaces

    Args:
        plot (:obj:`Plot3D`): plot

    Returns:
        :obj:`Report`: report with a dataset for each data generator of the surfaces
    """
    data_sets = {}
    for surface in plot.surfaces:
        data_sets[surface.x_data_generator.id] = DataSet(
            id=surface.x_data_generator.id,
            name=surface.x_data_generator.name,
            label=surface.x_data_generator.id,
            data_generator=surface.x_data_generator,
        )
        data_sets[surface.y_data_generator.id] = DataSet(
            id=surface.y_data_generator.id,
            name=surface.y_data_generator.name,
            label=surface.y_data_generator.id,
            data_generator=surface.y_data_generator,
        )
        data_sets[surface.z_data_generator.id] = DataSet(
            id=surface.z_data_generator.id,
            name=surface.z_data_generator.name,
            label=surface.z_data_generator.id,
            data_generator=surface.z_data_generator,
        )
    return Report(
        id=plot.id,
        name=plot.name,
        data_sets=sorted(data_sets.values(), key=lambda data_set: data_set.id))
